/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <math.h>
#include <strings.h>

#include "plugin.h"
#include "chain.h"
#include "xmalloc.h"


/** attempt to dlopen() an object file and populate an
    object_file_t for it.  returns 0 on success */
int init_object_file (object_file_t * object_file, char * filename) {
  const char * dlerr;
  
  object_file->filename = filename;
  
  /* open the object file */
  object_file->dl_handle = dlopen (filename, RTLD_NOW|RTLD_GLOBAL);
  if (!object_file->dl_handle) {
    fprintf (stderr, "%s: error opening shared object file '%s': %s\n",
             __FUNCTION__, filename, dlerror());
    return 1;
  }
  
  
  /* get the get_descriptor function */
  dlerror (); /* clear the error report */
  
  object_file->get_descriptor = (LADSPA_Descriptor_Function)
    dlsym (object_file->dl_handle, "ladspa_descriptor");
  
  dlerr = dlerror();
  if (dlerr) {
    fprintf (stderr, "%s: error finding ladspa_descriptor symbol: %s\n",
             __FUNCTION__, dlerr);
    dlclose (object_file->dl_handle);
    return 1;
  }
  
  
  /* get the plugin count.  this works because the index is always
     one less than the equivalent count */
  object_file->plugin_count = 0;
  while (object_file->get_descriptor (object_file->plugin_count)) {
    object_file->plugin_count++;
  }
  
  return 0;
}



/** pass all regular files in the specified dir to open_object_file
    to attempt to add them to the plugin_collection_t */
void collect_object_files_from_dir (plugin_collection_t * collection,
                                    const char * dir) {
  DIR * dir_stream;
  struct dirent * dir_entry;
  struct stat stat_info;
  char * file_name;
  int err;
  
  dir_stream = opendir (dir);
  if (!dir_stream) {
      fprintf (stderr, "%s: error opening directory '%s': %s\n",
               __FUNCTION__, dir, strerror (errno));
      return;
  }
  
  while ((dir_entry = readdir (dir_stream))) {
  
    /* we need to check the file type, which means we need
       a qualified file name */
    file_name = xmalloc (strlen (dir) + 1 +
                        strlen (dir_entry->d_name) + 1);
    
    strcpy (file_name, dir);
    if (file_name[strlen(dir) - 1] == '/') {
      strcpy (file_name + strlen(dir), dir_entry->d_name);
    } else {
      file_name[strlen(dir)] = '/';
      strcpy (file_name + strlen(dir) + 1, dir_entry->d_name);
    }
    
    /* stat the file */
    err = stat (file_name, &stat_info);
    if (err) {
      fprintf (stderr, "%s: could not stat file '%s': %s; skipping\n",
               __FUNCTION__, file_name, strerror (errno));
      free (file_name);
      continue;
    }
    
    /* check and create the entry */
    if (! S_ISREG (stat_info.st_mode)) {
      free (file_name);
    } else {
      object_file_t * object_file;
      
      object_file = xmalloc (sizeof (object_file_t));
      
      err = init_object_file (object_file, file_name);
      if (err) {
        free (file_name);
        free (object_file);
        continue;
      }
      
      object_file->next = collection->object_files;
      
      collection->object_files = object_file;
      collection->object_file_count++;
      
    }
  }
}

/** find the directories in LADSPA_PATH and use collect_object_files_from_dir
    to build the plugin_collection_t's list of object files */
void collect_object_files (plugin_collection_t * collection) {
  char * ladspa_path, * dir;
  
  ladspa_path = getenv ("LADSPA_PATH");
  if (!ladspa_path) {
    ladspa_path = "/usr/local/lib/ladspa:/usr/lib/ladspa";
  }
  
  dir = strtok (ladspa_path, ":");
  do {
    collect_object_files_from_dir (collection, dir);
  } while ((dir = strtok (NULL, ":")));

}

/** fill out all the port indicies
    and stuff like that */
int init_plugin (plugin_t * plugin,
                 object_file_t * object_file,
                 unsigned long index) {
  unsigned long port;
  unsigned int audio_in_count = 0;
  unsigned int audio_out_count = 0;
  int control_port_indicies_size = 1;


  /* create some port indicies */
  plugin->control_port_indicies = xmalloc (control_port_indicies_size *
                                          sizeof (unsigned long));
  
  plugin->object_file = object_file;
  plugin->index = index;
  plugin->descriptor = object_file->get_descriptor (index);
  plugin->name = plugin->descriptor->Name;
  plugin->control_port_count = 0;
  

  /* iterate through ports */  
  for (port = 0; port < plugin->descriptor->PortCount; port++) {


  
    if (LADSPA_IS_PORT_AUDIO (plugin->descriptor->PortDescriptors[port])) {
    
      /* set the port_indicies for these */
    
      if (LADSPA_IS_PORT_INPUT (plugin->descriptor->PortDescriptors[port])) {
      
        switch (audio_in_count) {
          case 0:
            plugin->port_indicies[IN_R] = plugin->port_indicies[IN_L] = port;
            break;
          case 1:
            plugin->port_indicies[IN_R] = port;
            break;
          case 2:
            /* not a stereo or mono plugin */
            free (plugin->control_port_indicies);
            return 1;
        }
        audio_in_count++;
        
      } else {

        switch (audio_out_count) {
          case 0:
            plugin->port_indicies[OUT_R] = plugin->port_indicies[OUT_L] = port;
            break;
          case 1:
            plugin->port_indicies[OUT_R] = port;
            break;
          case 2:
            /* not a stereo or mono plugin */
            free (plugin->control_port_indicies);
            return 1;
        }
        audio_out_count++;

      }
      
    } else {
      /* we're not concerned with control output ports */
      if (LADSPA_IS_PORT_OUTPUT (plugin->descriptor->PortDescriptors[port])) {
        continue;
      }
 
      
      /* check if we're running out of memory in the control port indicies */
      if (plugin->control_port_count + 1 > control_port_indicies_size) {
        unsigned long * tmp;
/*        unsigned long j;
        
        printf ("%s: plugin: %s, port: %ld\n", __FUNCTION__, plugin->name, port);
        for (j = 0; j < plugin->control_port_count; j++) {
          printf ("%s: before copy, control port at index %ld\n",
                  __FUNCTION__, plugin->control_port_indicies[j]);
        }*/
        
        tmp = plugin->control_port_indicies;
        
        plugin->control_port_indicies = xmalloc (control_port_indicies_size * 2 *
                                                 sizeof (unsigned long));
        
        memcpy (plugin->control_port_indicies, tmp,
                control_port_indicies_size * sizeof (unsigned long));
        
        control_port_indicies_size *= 2;
        
        free (tmp);

/*        for (j = 0; j < plugin->control_port_count; j++) {
          printf ("%s: after copy, control port at index %ld\n",
                  __FUNCTION__, plugin->control_port_indicies[j]);
        }*/
        
      }

      
      plugin->control_port_indicies[plugin->control_port_count] = port;
      plugin->control_port_count++;
    }
  }
  
  /* bad i/o configurations */
  if (!audio_in_count || !audio_out_count || audio_in_count != audio_out_count) {
    free (plugin->control_port_indicies);
    return 1;
  }
  
  /* stereo plugin? */  
  if (audio_in_count == 2) { 
    plugin->mono = FALSE;
  } else {
    plugin->mono = TRUE;
  }
  
  return 0;
}


void create_plugins (plugin_collection_t * collection) {
  object_file_t * object_file;
  unsigned long plugin_index;
  plugin_t * plugin;
  int err;
  
  for (object_file = collection->object_files;
       object_file;
       object_file = object_file->next) {
    for (plugin_index = 0;
         plugin_index < object_file->plugin_count;
         plugin_index++) {
         
      plugin = xmalloc (sizeof (plugin_t));
      
      err = init_plugin (plugin, object_file, plugin_index);
      if (err) {
        free (plugin);
      } else {
        plugin->next = collection->plugins;
        collection->plugins = plugin;
        collection->plugin_count++;
      }
      
    }
  }
}

/* sort alphabetically */
void sort_collection (plugin_collection_t * collection) {
  gboolean changed = TRUE;
  plugin_t * plugin, * last_plugin, * tmp;
  int cmp;
  
  if (!collection->plugins->next) {
    return;
  }
  
  while (changed) {
    
    changed = FALSE;
    last_plugin = NULL;
    
    for (plugin = collection->plugins;
         plugin && plugin->next;
         plugin = plugin->next) {

      cmp = strcasecmp (plugin->name, plugin->next->name);

      if (cmp > 0) {

        tmp = plugin->next->next;
        if (last_plugin) {
          last_plugin->next = plugin->next;
          plugin->next->next = plugin;
        } else {
          collection->plugins = plugin->next;
          plugin->next->next = plugin;
        }
        plugin->next = tmp;
        
        changed = TRUE;
        
        if (last_plugin) {
          plugin = last_plugin->next;
        } else {
          plugin = collection->plugins->next;
        }

      }
      
      last_plugin = plugin;
    }
  }
}

void print_collection (plugin_collection_t * collection) {
  plugin_t * plugin;
  
  for (plugin = collection->plugins;
       plugin;
       plugin = plugin->next) {
    printf ("%s: plugin: %s (%s, %s)\n", __FUNCTION__,
            plugin->name,
            plugin->object_file->filename,
            (plugin->mono ? "mono" : "stereo"));
  }
}

void create_plugin_array (plugin_collection_t * collection) {
  unsigned long i = 0;
  plugin_t * plugin;
  
  collection->plugin_array = xmalloc (sizeof (plugin_t *) *
                                      collection->plugin_count);

  for (plugin = collection->plugins;
       plugin;
       plugin = plugin->next) {
    collection->plugin_array[i] = plugin;
    plugin->collection_index = i;
    
    i++;
  }
}

LADSPA_Data get_default_control_value (const LADSPA_PortRangeHint * hint) {
  LADSPA_Data upper, lower;
  
  /* set upper and lower, possibly adjusted to the sample rate */
  if (LADSPA_IS_HINT_SAMPLE_RATE(hint->HintDescriptor)) {
    upper = hint->UpperBound * (LADSPA_Data) sample_rate;
    lower = hint->LowerBound * (LADSPA_Data) sample_rate;
  } else {
    upper = hint->UpperBound;
    lower = hint->LowerBound;
  }

  if (LADSPA_IS_HINT_HAS_DEFAULT(hint->HintDescriptor)) {
      
           if (LADSPA_IS_HINT_DEFAULT_MINIMUM(hint->HintDescriptor)) {
    
      return lower;
       
    } else if (LADSPA_IS_HINT_DEFAULT_LOW(hint->HintDescriptor)) {
        
      if (LADSPA_IS_HINT_LOGARITHMIC(hint->HintDescriptor)) {
        return exp(log(lower) * 0.75 + log(upper) * 0.25);
      } else {
        return lower * 0.75 + upper * 0.25;
      }

    } else if (LADSPA_IS_HINT_DEFAULT_MIDDLE(hint->HintDescriptor)) {
        
      if (LADSPA_IS_HINT_LOGARITHMIC(hint->HintDescriptor)) {
        return exp(log(lower) * 0.5 + log(upper) * 0.5);
      } else {
        return lower * 0.5 + upper * 0.5;
      }

    } else if (LADSPA_IS_HINT_DEFAULT_HIGH(hint->HintDescriptor)) {
      
      if (LADSPA_IS_HINT_LOGARITHMIC(hint->HintDescriptor)) {
        return exp(log(lower) * 0.25 + log(upper) * 0.75);
      } else {
        return lower * 0.25 + upper * 0.75;
      }
      
    } else if (LADSPA_IS_HINT_DEFAULT_MAXIMUM(hint->HintDescriptor)) {
      
      return upper;
    
    } else if (LADSPA_IS_HINT_DEFAULT_0(hint->HintDescriptor)) {
      
      return 0.0;
      
    } else if (LADSPA_IS_HINT_DEFAULT_1(hint->HintDescriptor)) {
      
      if (LADSPA_IS_HINT_SAMPLE_RATE(hint->HintDescriptor)) {
        return (LADSPA_Data) sample_rate;
      } else {
        return 1.0;
      }
      
    } else if (LADSPA_IS_HINT_DEFAULT_100(hint->HintDescriptor)) {
      
      if (LADSPA_IS_HINT_SAMPLE_RATE(hint->HintDescriptor)) {
        return 100.0 * (LADSPA_Data) sample_rate;
      } else {
        return 100.0;
      }
      
    } else if (LADSPA_IS_HINT_DEFAULT_440(hint->HintDescriptor)) {
      
      if (LADSPA_IS_HINT_SAMPLE_RATE(hint->HintDescriptor)) {
        return 440.0 * (LADSPA_Data) sample_rate;
      } else {
        return 440.0;
      }
      
    }  
      
  } else { /* try and find a reasonable default */
        
           if (LADSPA_IS_HINT_BOUNDED_BELOW(hint->HintDescriptor)) {
      return lower;
    } else if (LADSPA_IS_HINT_BOUNDED_ABOVE(hint->HintDescriptor)) {
      return upper;
    }
  }

  return 0.0;
}

void create_default_plugin_settings (plugin_collection_t * collection) {
  plugin_t * plugin;
  unsigned long i, j;
  
  collection->default_plugin_settings =
    xmalloc (sizeof (LADSPA_Data *) * collection->plugin_count);
  
  for (i = 0; i < collection->plugin_count; i++) {
    plugin = collection->plugin_array[i];
    collection->default_plugin_settings[i] =
      xmalloc (sizeof (LADSPA_Data) * plugin->control_port_count);
    
    for (j = 0; j < plugin->control_port_count; j++) {
      collection->default_plugin_settings[i][j] =
        get_default_control_value (plugin->descriptor->PortRangeHints + j);
    }
  }
}

LADSPA_Data ** duplicate_default_plugin_settings (plugin_collection_t * collection) {
  LADSPA_Data ** settings;
  unsigned int i, j;
  
  settings = xmalloc (sizeof (LADSPA_Data *) * collection->plugin_count);
  
  for (i = 0; i < collection->plugin_count; i++) {
  
    settings[i] = xmalloc (sizeof (LADSPA_Data) *
                           collection->plugin_array[i]->control_port_count);
    
    for (j = 0; j < collection->plugin_array[i]->control_port_count; j++) {
      settings[i][j] = collection->default_plugin_settings[i][j];
    }
    
  }
  
  return settings;
}

unsigned long find_plugin (plugin_t * plugin,
                           plugin_collection_t * collection) {
  unsigned long i = 0;
  plugin_t * cplugin;
  
  cplugin = collection->plugins;
  
  while (plugin != cplugin) {
    i++;
    cplugin = cplugin->next;
    if (!cplugin) {
      fprintf (stderr, "%s: couldn't find a specified plugin; major badness going on\n", __FUNCTION__);
      fprintf (stderr, "%s: I'm going to pretend that plugin 0 was it. This probably won't work, so expect a crash\n", __FUNCTION__);
      return 0;
    }
  }

  return 0;
}

void init_plugin_collection (plugin_collection_t * collection) {
  collection->object_files = NULL;
  collection->object_file_count = 0;
  
  collect_object_files (collection);
  
  if (!collection->object_files) {
    fprintf (stderr, "%s: no plugins; exiting. check your LADSPA_PATH environment variable\n",
             __FUNCTION__);
    exit (1);
  }
  
  collection->plugins = NULL;
  collection->plugin_count = 0;
  create_plugins (collection);
  
  if (!collection->plugins) {
    fprintf (stderr, "%s: no plugins; exiting. check your LADSPA_PATH environment variable\n",
             __FUNCTION__);
    exit (1);
  }
  
  sort_collection (collection);
  
  create_plugin_array (collection);
  
  create_default_plugin_settings (collection);
  
  return;
}

void destroy_plugin_collection (plugin_collection_t * collection) {
  unsigned long i;
  plugin_t * plugin;
  object_file_t * object_file, * next_object_file;
  int err;
  
  for (i = 0; i < collection->plugin_count; i++) {
  }
  
  for (i = 0; i < collection->plugin_count; i++) {
    /* free plugins */
    plugin = collection->plugin_array[i];
    free (plugin->control_port_indicies);
    free (plugin);

    /* free plugin settings */
    free (collection->default_plugin_settings[i]);
  }
  free (collection->plugin_array);
  free (collection->default_plugin_settings);
  
  /* close and free all the object files */
  for (object_file = collection->object_files;
       object_file;
       object_file = next_object_file) {
    next_object_file = object_file->next;
    
    err = dlclose(object_file->dl_handle);
    if (err) {
      fprintf (stderr, "%s: error closing plugin object '%s': %s",
               __FUNCTION__, object_file->filename, dlerror());
    }
    
    free (object_file->filename);
    free (object_file);
  }
}
